#pragma once

#include "exportrules.h"
#include "asm_defines.h"
#include "codebook.h"
#include <utility>


class ProgressCallback;

class SSTD_API SOM {
public:
	using tag_map_t = std::unordered_map<const Vector *, std::unordered_map<unsigned int, unsigned int> >;
	const Vector * findClosest(const Vector & vec ) const;
private:

	unsigned int rows, cols;
	Vector * nodes;
	ProgressCallback * p_cb;
	
	void train( const Vector & vec, unsigned int t );
	void update( Vector & m, const Vector & x, unsigned int t, unsigned long dst );
	
	double alpha( unsigned int t );
	unsigned int neighborhood_radius( unsigned int t );
	double d_decay(unsigned long d, unsigned int t);
	
	void visualize_grid(const tag_map_t & all_tags, const std::vector<std::string> & tag_strings) const;
	
	
	
public:
	SOM() : rows(0), cols(0), nodes(nullptr), p_cb(nullptr) {}
	SOM( unsigned int node_rows, unsigned int node_cols, ProgressCallback * p_cb = nullptr);
	SOM(const SOM & rhs) : SOM(rhs.rows, rhs.cols, rhs.p_cb) {
	    for (unsigned int i = 0; i < rhs.rows*rhs.cols; i++) {
	        this->nodes[i] = rhs.nodes[i];
	    }
	}	
	SOM(SOM && rhs) : rows(rhs.rows), cols(rhs.cols), p_cb(rhs.p_cb), nodes(rhs.nodes) {
		rhs.rows = 0;
		rhs.cols = 0;
		rhs.nodes = nullptr;
		rhs.p_cb = nullptr;
	}
	~SOM() { delete [] nodes; }
	
	SOM& operator=(const SOM & rhs);	
	SOM& operator=(SOM && rhs) noexcept {
		using std::swap;
		swap(this->rows, rhs.rows);
		swap(this->cols, rhs.cols);
		swap(this->nodes, rhs.nodes);
		swap(this->p_cb, rhs.p_cb);
		
		return *this;
	}
	
	bool initialized() const { return nodes != nullptr; }
	
	/** Trains the SOM, any tags in the training data are ignored. 
	 * if bootstrap == true, the algorithm will initiallize the SOM-map with randomly 
	 * drawn vectors from the training data. Otherwise it will leave the data as-is. 
	 **/
	void train(const Codebook & training_data, unsigned int iterations );
	
	/** Uses the tagging data to tag the codewords in the codebook. 
	 *  If any data is node is left without any hit, it is dropped from the returned 
	 *  Codebook.
	 **/
	Codebook tag_and_purge( const Codebook & tagging_data ) const;

	using const_iterator = const Vector *;
	using iterator       = const Vector *; 
	const_iterator begin() const { return nodes; }
	const_iterator end() const { return nodes + rows*cols; }

};

