#pragma once
#include "exportrules.h"
#include <utility>
#include <cmath>
#include <istream>
#include <ostream>
#include <cstdint>
#include <string>
#include <cassert>
#include <cstring>

template<class T>
class SSTD_API VectorT {
public:
	using ElemType = T;

private:
	T * data;
	uint32_t  _size;
	
	static inline T * alloc(uint32_t _size) {
		return _size ? new T[_size] : nullptr;
	}
	
	static inline void dealloc( T * ptr ) {
		delete [] ptr;
	}
	
public:
	VectorT() : data(nullptr), _size(0) {}

	explicit VectorT(uint32_t _size, T fill = 0 ) : data(alloc(_size)), _size(_size) {
		for ( uint32_t i = 0; i < _size; i++) {
			data[i] = fill;
		}
	}

	VectorT( VectorT && rhs ) : VectorT() {
		this->data = rhs.data;
		this->_size = rhs._size;
		rhs.data = nullptr;
		rhs._size = 0;
	}

	VectorT& operator=(VectorT && rhs ) {
		swap(*this, rhs);
	}

	VectorT( const VectorT & rhs) : data(alloc(rhs._size)), _size(rhs._size) {
		for ( uint32_t i = 0; i < _size; i++) {
			this->data[i] = rhs.data[i];
		}
	}
		
	VectorT& operator=(const VectorT& rhs ) {
		if (rhs._size > this->_size){
			T * temp = alloc(rhs._size);
			dealloc(this->data);
			data = temp;
			_size = rhs._size;
		}
		
		memcpy(data, rhs.data, _size*sizeof(T));
	
		return *this;
	}
	
	virtual ~VectorT() {
		dealloc(data);
	}

	uint32_t size() const                  { return _size;   }
	T  operator[]( unsigned int i ) const  { return data[i]; }
	T& operator[]( unsigned int i )        { return data[i]; }

	T norm_1() const {
		using std::abs;
		T out = 0;
		for ( uint32_t i = 0; i < _size; i++) {
			out += abs(data[i]);
		}
	
		return out;
	}

	double length() const { norm_2(nullptr); }
	double norm_2(T * sumSquared_out) const ;
    T norm_inf() const;
	
	VectorT& operator+=(const VectorT& rhs);
	VectorT& operator-=(const VectorT& rhs);
	
	VectorT operator-() const;
	
	template<class scalarT>
	VectorT& operator*=(const scalarT & rhs ) {
		for (uint32_t i = 0; i < _size; i++) {
			data[i] *= rhs;
		}
		return *this;
	}
	
	template<class T2>
	VectorT<T2> convertTo() const {
	    VectorT<T2> temp(size());
	    for (int i = 0; i < size(); i++) {
	        temp[i] = static_cast<T2>(this->data[i]);
	    }
	    
	    return std::move(temp);
	}
	
	/** If readFrom or write to returns false. The VectorT may be in an invalid state. */
	bool readFrom(std::istream & in);
	bool writeTo(std::ostream & out) const;
	
	friend SSTD_API inline void swap( VectorT & lhs, VectorT & rhs) noexcept {
		using std::swap;
		swap( lhs.data, rhs.data);
		swap( lhs._size, rhs._size);
	}

	using iterator = T*;
	using const_iterator = const T*;

	iterator begin() { return data; }
	iterator end()   { return data + _size; }

	const_iterator begin() const { return data; }
	const_iterator end()   const { return data + _size; }
};

template<class T>
double VectorT<T>::norm_2(T * sumSquared_out) const {
	T sumSquared = 0;
	for ( uint32_t i = 0; i < _size; i++) {
		sumSquared += data[i]*data[i];
	}

	if (sumSquared_out)
		*sumSquared_out = sumSquared;

	return sqrt(sumSquared);
}

template<class T>
T VectorT<T>::norm_inf() const {
	T candidate = abs(data[0]);
	for ( uint32_t i = 1; i < _size; i++) {
		register T temp = abs(data[i]);
		if (temp > candidate) {
			candidate = temp;
		}
	}
	return candidate;
}

template<class scalarT, class T> 
SSTD_API VectorT<T> operator*(const VectorT<T> & lhs, const scalarT & rhs ) {
	return std::move( VectorT<T>(lhs) *= rhs );	
}

template<class scalarT, class T> 
SSTD_API VectorT<T> operator*(const scalarT & lhs, const VectorT<T> & rhs) {
	VectorT<T> temp(rhs);
	
	for ( int i = 0; i < temp.size(); i++) {
		temp[i] = lhs*temp[i];
	}
	
	return std::move(temp);	
}

template<class T>
static inline VectorT<T> operator-( const VectorT<T> & lhs, const VectorT<T> & rhs ) {
	return std::move(VectorT<T>(lhs) -= rhs);
}

template<class T>
static inline VectorT<T> operator+( const VectorT<T> & lhs, const VectorT<T> & rhs ) {
	return std::move(VectorT<T>(lhs) += rhs);
}

template<class T>
T SSTD_API d_2( const VectorT<T> & lhs, const VectorT<T> & rhs ) { 
	assert(lhs.size() == rhs.size());
	assert(lhs.size() > 0);
	assert(rhs.size() > 0);
	
	T temp = 0;
	for (int i = 0; i < lhs.size(); i++) {
		T diff = rhs[i]-lhs[i];
		temp += diff*diff;
	}
	return temp;
}

template<class T>
static inline double d_euclide( const VectorT<T> & lhs, const VectorT<T> & rhs ) {
	using std::sqrt;
	return sqrt(d_2(lhs, rhs));		
}

template<class T>
T SSTD_API dot( const VectorT<T> & lhs, const VectorT<T> & rhs ) {
	using std::min;
	const uint32_t N = min(lhs.size(), rhs.size());
	T ret = 0;
	for (uint32_t i = 0; i < N; i++) {
		ret += lhs[i]*rhs[i];
	}
	return ret;
}

template<class T>
T SSTD_API d_inf( const VectorT<T> & lhs, const VectorT<T> & rhs ) {
	assert(lhs.size() == rhs.size());
	assert(lhs.size() > 0);
	assert(rhs.size() > 0);
	
	unsigned int candidate = abs(lhs[0]-rhs[0]);
	
	for (uint32_t i = 1; i < lhs.size(); i++) {
		register unsigned int temp = abs(lhs[i]-rhs[i]);
		if (temp > candidate) {
			candidate = temp;
		}
	}
	
	return candidate;
}


// Implementations.
template<class T>
bool VectorT<T>::readFrom(std::istream & in) {
	// Assume there will be a VectorT next on the stream.
	uint32_t temp_size;
	bool notFail = in.read( reinterpret_cast<char*>(&temp_size), sizeof(temp_size));
	
	if (notFail && temp_size != _size) {
		T * temp_data = alloc(temp_size);
		dealloc(data);
		data = temp_data;
		_size = temp_size;
	}
	
	return notFail && in.read( reinterpret_cast<char*>(data), _size*sizeof(T));
}

template<class T>
bool VectorT<T>::writeTo(std::ostream & out) const {
	return out.write( reinterpret_cast<const char*>(&_size), sizeof(_size))
		&& !!(out.write( reinterpret_cast<const char*>(data), _size*sizeof(T)));
}



template<class T>
VectorT<T>& VectorT<T>::operator+=(const VectorT<T>& rhs) {
	if (rhs._size > this->_size) {
		VectorT<T> temp(rhs);
		for (uint32_t i = 0; i < _size; i++) {
			temp.data[i] += this->data[i];
		}
		swap(*this, temp);
	}
	else {
		using std::min;
		const uint32_t N = min(this->_size, rhs._size);
		for (uint32_t i = 0; i < N; i++) {
			data[i] += rhs.data[i];
		}
	}
	return *this;
}

template<class T>
VectorT<T>& VectorT<T>::operator-=(const VectorT<T>& rhs) {
	if (rhs._size > this->_size) {
		VectorT<T> temp(std::move(-rhs));
		for (uint32_t i = 0; i < _size; i++) {
			temp.data[i] += this->data[i];
		}
		swap(*this, temp);
	}
	else {
		using std::min;
		const uint32_t N = min(this->_size, rhs._size);
		for (uint32_t i = 0; i < N; i++) {
			data[i] -= rhs.data[i];
		}
	}
	return *this;
}

template<class T>
VectorT<T> VectorT<T>::operator-() const {
	VectorT temp(*this);
	for (uint32_t i = 0; i < temp._size; i++) {
		temp.data[i] = -temp.data[i];
	}
	return std::move(temp);
}

template<class T>
std::string to_string(const VectorT<T> & obj ) {
	using std::to_string;
	std::string s;
	for ( int i = 0; i < obj.size(); i++) {
		 s += to_string(obj[i]) + (i != obj.size()-1 ? ", " : "");
	}
	
	return s;
}

template<class T>
SSTD_API std::ostream &  operator<<( std::ostream & out, const VectorT<T> & obj ) {
	out << "[ ";
	for ( int i = 0; i < obj.size(); i++) {
		out << obj[i] << ' ';
	}
	out << ']';
	return out;		
}

using Vector = VectorT<float>;

using fVec   = VectorT<float>;
using iVec   = VectorT<int>;
using dVec   = VectorT<double>;
using bVec   = VectorT<char>;
using ubVec  = VectorT<unsigned char>;


