#include "SOM.h"
#include <cstdlib>
#include <atomic>
#include <algorithm>
#include <unordered_map>
#include <vector>
#include <unistd.h>
#include <iostream> // REMOVE

#include <thread>
#include "progress_callback.h"

using namespace std;

const int NEIGHBORHOOD_DECAY_LIMIT = 100;
const long NEIGHBORHOOD_DECAY_MIN  = 0;

/////////////////////////
//   TO BE TWEAKED     //
/////////////////////////

inline double exp_decay(double t , double init, double halftime ) {
	const double B = -log(2)/abs(halftime);
	return init*exp(B*t);
}

double SOM::alpha( unsigned int t ) {
	const double halftime = 50;
	return exp_decay(t, 0.8, halftime);
}

double SOM::d_decay(unsigned long d, unsigned int t) {
	unsigned int nh_radius = neighborhood_radius(t);
	return nh_radius ? max( 1.0 - static_cast<double>(d)/nh_radius, 0.0) : 1.0 ;
}

unsigned int SOM::neighborhood_radius( unsigned int t ) {
	const long m = max(rows,cols);
	
	return max( NEIGHBORHOOD_DECAY_MIN, m - static_cast<long>(m*t)/NEIGHBORHOOD_DECAY_LIMIT ); // Decrease monotonically for NEIGHBORHOOD_DECAY_LIMIT iterations.
}

/////////////////////////
//  END TO BE TWEAKED  //
/////////////////////////

SOM::SOM( unsigned int node_rows, unsigned int node_cols, ProgressCallback * p_cb  ) 
: rows(node_rows)
, cols(node_cols)
, nodes( new Vector[node_rows*node_cols] ) 
, p_cb(p_cb)
{}

SOM& SOM::operator=(const SOM & rhs) {
	    if ( rows*cols < rhs.rows*rhs.cols ){
	        Vector * temp = new Vector[rhs.rows*rhs.cols];
			
			try {
	        	for (unsigned int i = 0; i < rhs.rows*rhs.cols; i++) {
	        		temp[i] = rhs.nodes[i];
	        	}
	  		} catch(...) {
	  		    delete [] temp;
	  		    throw;
	  		}
			
			delete [] this->nodes;
	        this->nodes = temp;
	    }
	    else {
	        for (unsigned int i = 0; i < rhs.rows*rhs.cols; i++) {
   		        this->nodes[i] = rhs.nodes[i];
	  		}
	    }
	    
	    this->rows = rhs.rows;
		this->cols = rhs.cols;
		this->p_cb = rhs.p_cb;
	    
	    return *this;
	}


const Vector * SOM::findClosest(const Vector & vec ) const {
	const Vector * candidate = nodes;
	unsigned long d = d_2(*candidate, vec);
	
	for ( int i = 0; i < rows*cols; i++) {
		unsigned long di = d_2(nodes[i], vec);
		if ( di < d ){
			candidate = nodes + i;
			d = di;
		}
	}
	return candidate;
}

	/** Trains the SOM, any tags in the training data are ignored. 
	 * if bootstrap == true, the algorithm will initiallize the SOM-map with randomly 
	 * drawn vectors from the training data. Otherwise it will leave the data as-is. 
	 **/
void SOM::train(const Codebook & training_data, unsigned int iterations ) {
	
//	cout << "Internal nodes: ";
	// Always bootstrap
	for ( int i = 0; i < rows*cols; i++) {
		nodes[i] = training_data[ rand() % training_data.size() ];
		//cout << nodes[i] << ", ";
	}
	//cout << endl;
	

	
	for (unsigned int i = 0; i < iterations; i++ ) {
		if (p_cb) { p_cb->updateProgress(static_cast<double>(i*100)/iterations, "Training: ");}
		train( training_data[ rand() % training_data.size() ], i );
	}
	
	if (p_cb) { p_cb->updateProgress(100, "Training: ");}
}

void SOM::update( Vector & m, const Vector & x, unsigned int t, unsigned long distance ) {
	m += d_decay(distance,t)*alpha(t)*(x - m);
}

void SOM::train( const Vector & vec, unsigned int t ) {
	const Vector * closest_ptr = findClosest( vec );
	const int row = (closest_ptr-nodes)/cols;
	const int col = (closest_ptr-nodes)%cols;

	int R =  neighborhood_radius(t); // Neighbourhood radius.

	const int dr_min = -min(row, R);
	const int dr_max = min<int>(rows-row-1, R);
	
	const Vector closest = *closest_ptr; // Keep copy at time t.
	
	// Hexagonal grid update.
	for (int dr = dr_min; dr <= dr_max; dr++) {
		const int dr_abs = abs(dr);

		int c_min = col - (R - dr_abs);
		if (dr_abs & 1) { // if this row has a different parity than the middle row.
			c_min += (dr_abs & 1) - 1;
		}

		int c_max = c_min + 2*R - dr_abs;

		// Clamp min and max to gridsize
		c_min = max(0, c_min);
		c_max = min<int>(cols-1, c_max);


		for (int c = c_min; c <= c_max; c++ ) {
			int r = row + dr;
			
			//cout << "  (" << row << ", " << col << ") --- (" << r << ", " << c <<")    dr= " << dr << endl;
			update( nodes[r*cols + c], vec, t, dr_abs );
		}
	}
}



int tag(SOM::tag_map_t * tag_map, const SOM * som, const Codebook * tagging_data, const int begin, const int end,  std::atomic<int> * counter) {
	// Populate tag map.
	for ( int i = begin; i < end;  i++ ) {
		auto & cw = (*tagging_data)[i];
		(*tag_map)[som->findClosest( cw )][cw.label()]++;		
		++(*counter);
	}
	return 0;
}

void merge(SOM::tag_map_t & lhs, const SOM::tag_map_t & rhs) {
	for (auto & entry : rhs) {
		for ( auto & tag : entry.second ) {
			lhs[entry.first][tag.first] += tag.second;
		}
	}
}

int out(double f ) {
	return sin(f);
}

Codebook SOM::tag_and_purge( const Codebook & tagging_data ) const {

	const int n_threads = std::thread::hardware_concurrency();
	
	tag_map_t tag_maps[n_threads];
	
	std::vector<std::thread> threads;
	
	int partition = tagging_data.size()/n_threads;
	
	atomic<int> progress_counter(0);
	
	for (int i = 0; i < n_threads-1; i++) {
		threads.emplace_back(std::thread(tag, tag_maps + i, this, &tagging_data, i*partition, (i+1)*partition, &progress_counter ));
	}
	threads.emplace_back(std::thread(tag, tag_maps + n_threads-1, this, &tagging_data, (n_threads-1)*partition, tagging_data.size(), &progress_counter ));
	
	if (p_cb) {
		while (progress_counter < tagging_data.size()) {
			p_cb->updateProgress(static_cast<double>(progress_counter*100)/tagging_data.size(), " Tagging: ");
			usleep(1000);
		}
		p_cb->updateProgress(100, " Tagging: ");
	}
	
	for (int i = 0; i < threads.size(); i++) {
		threads[i].join();
		if ( i > 0) {
			merge(tag_maps[0], tag_maps[i]);
		}
	}
	
	Codebook ret;

	// Map integer tag to string tag
	std::vector<string> tags(tagging_data.getLabelMap().size());
	for (auto & t : tagging_data.getLabelMap() ) {
		tags[t.second] = t.first;
	}
	
	int purged = 0;
	int i = 0;
	for ( auto & entry : tag_maps[0] ) {
		if ( entry.second.size() == 0 ) {
			purged++;
		}
		else {
			
			//cout << "Vector " << entry.first << " - " << *entry.first << "\n";
			int tag = entry.second.begin()->first;
			int n   = entry.second.begin()->second;
			for ( auto & tag_num : entry.second ) {
				//cout << "\t" << tags[tag_num.first ] << ": " << tag_num.second << endl;
				if ( tag_num.second > n ) {
					tag = tag_num.first;
					n = tag_num.second;
				}
			}
			
			ret.addCodeword(*(entry.first), tags[tag] );
		}	
	}
	
	//visualize_grid(tag_maps[0], tags);
	
	return ret;
}

#include <iomanip>
#include <sstream>

void SOM::visualize_grid(const tag_map_t & all_tags, const vector<string> & tag_strings) const {
	

	for ( int r = 0; r < rows; r++ ){
		cout << (r&1 ? "        " : "");
		for (int c = 0; c < cols; c++) {
			auto elem_ptr = nodes + r*cols + c;
			auto tags_it = all_tags.find(elem_ptr);
			ostringstream sout;
			sout << (tags_it != all_tags.end()? "*" : "") << (void*)(0xffffLL & (long long)(elem_ptr));
			cout << std::setw(8) << sout.str();
		}
		cout << endl;
	}
}

            
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	

