#include <iostream>
#include <algorithm>
#include <cstdlib>
#include <ctime>
#include "lvq.h"
#include "progress_callback.h"
#include "SOM.h"
#include "codebook_reduce.h"

#include <iomanip>
#include <thread>
#include <atomic>
#include <unistd.h>

#include <fstream>

using namespace std;

int print_statistics(const Codebook & cb_in) {
	
		cout << "Labels: \n";

		vector<pair<string, int>> labels(cb_in.getLabelMap().size());

		int maxlen = 0;
		for (auto label : cb_in.getLabelMap()) {
			labels[label.second] = make_pair( label.first, 0);
			if (label.first.size() > maxlen) {
				maxlen = label.first.size();
			}
		}

		for ( auto cw : cb_in ) {
			labels[cw.label()].second++;			
		}
			
		for ( auto entry : labels ) {
			cout << setw(maxlen + 4) << entry.first << " :  " << entry.second << '\n';
		}
		cout << flush;
	
	return 0;
}


void test_1( const VQ & encoder, const Codebook & training_data, const std::vector<unsigned int> &label_map, const int begin, const int end, atomic<int> & hits, atomic<int> & counter ) {
	for (int i = begin; i < end; i++ ) { 
		auto & elem = training_data[i];
		if (label_map[elem.label()] == encoder.encode(elem) )
			hits++;
		counter++;
	}
}

int test(const VQ & encoder, const Codebook & training_data) {
	std::vector<thread> threads;
	atomic<int> hits(0), counter(0);
	
	int nthreads = 1;//std::thread::hardware_concurrency();
	int part 	 = training_data.size()/nthreads;
	
	std::vector<unsigned int> label_map(training_data.getLabelMap().size());
	for ( unsigned int i = 0; i < label_map.size(); i++) {
		label_map[i] = encoder.getLabelId( training_data.getLabelString(i) );
	}
	
	for ( int i = 0; i < nthreads-1; i++) {
		threads.emplace_back( thread(test_1, std::ref(encoder), std::ref(training_data), std::ref(label_map), i*part, (i+1)*part, std::ref(hits), std::ref(counter)));
	}
	threads.emplace_back( thread(test_1, std::ref(encoder), std::ref(training_data), std::ref(label_map), (nthreads-1)*part, training_data.size(), std::ref(hits), std::ref(counter)));
	
	ProgressCallback p_cb;
	while (counter < training_data.size()) {
		p_cb.updateProgress(static_cast<double>(counter*100)/training_data.size(), " Testing: ");
		usleep(1000);
	}
	p_cb.updateProgress(100, " Testing: ");
	
	
	for ( auto & th : threads ) {
		th.join();
	}
	return hits;
}


int main() {

	srand(time(0));
	ofstream fout;

#if 1
	
	// Load and reduce Inputdata.
	Codebook zero = loadCodebook("zero.cb");
	Codebook one = loadCodebook("one.cb");
	
	std::vector<const Codebook *> cbs;
	cbs.push_back(&zero);
	cbs.push_back(&one);
	
	auto components = find_significant_components(cbs);
	
	fout.open("components.comp");
	
	for ( auto c : components ) {
	   fout << c << " ";
	}
	fout << endl;
	fout.close();
	
	Codebook training_data;
	training_data.combine(zero);
	training_data.combine(one);
	
	for (auto & elem : training_data ) {
	    elem = reduce_dimensions(elem, components);
	}
	
	// CREATE SOM:
	ProgressCallback pcb;	
	Codebook result;
	SOM map(3, 3, &pcb);
	
	map.train(training_data, 100*training_data.size());
	
	LVQ lvq(map.tag_and_purge(training_data));
	
	print_statistics(lvq);
	
	int hits = test(lvq, training_data);
	cout << "SOM: HitRatio: " << setprecision(5) << static_cast<double>(hits)/training_data.size() << endl;
	
	lvq.train_vq(training_data, 10000);
	
	saveCodebook( "trained_lvq.cb", lvq);
	
	hits = test(lvq, training_data);
	cout << "LVQ: HitRatio: " << setprecision(5) << static_cast<double>(hits)/training_data.size() << endl;

#else
    LVQ lvq( loadCodebook("trained_lvq.cb"));
    
    std::vector<unsigned int> components;
    ifstream fin("components.comp");
    
    unsigned int temp;
    while (fin >> temp) {
        components.push_back(temp);
    } 
    
    fin.close();
#endif


	Codebook testset = loadCodebook("alt.cb");
	
	for (auto & elem : testset ) {
	    elem = reduce_dimensions(elem, components);
	}
	
	std::vector<int> string_labels(lvq.getLabelMap().size());
	
	for( auto & entry: lvq.getLabelMap()) {
	   if (entry.first == "ONE") string_labels[ entry.second ] = 1;
	   else string_labels[ entry.second ] = 0;
	}
	
	unsigned int N = 200;
	
	for (int i = 0; i < N; i++ ) {
		cout << (0x1 & (i ^ 1));
	}
	cout << endl;
	
	for (int i = 0; i < N; i++ ) {
		cout << string_labels[lvq.encode( testset[i] )];
	}
	cout << endl;
	
	for (int i = 0; i < N; i++ ) {
		cout << ((unsigned int)(0x1 & (i ^ 1)) != string_labels[lvq.encode( testset[i] )] ? "^" : " ");
	}
	
	cout << endl;
	
}




