#include "options.h"
#include <openssl/bn.h>
#include <getopt.h>
#include <stdlib.h>

#define BN_INIT(bn) bn = BN_new(); BN_init(bn)

#define DEFAULT_KEY_BITS 1024

static RSA * training(int zo);
static RSA * train_none();

// Available options
enum {
	PRIME_P = 'P',
	PRIME_Q = 'Q',
	KEYBITS = 's',
	TRAIN_0 = '0',
	TRAIN_1 = '1',
	TRAIN_NONE = 'N'
};

const char * OPTIONS = "P:Q:s:01N";

static RSA* loadRSA(const char * P, const char *Q, int keybits ) {
	RSA * rsa = RSA_new();
	BN_CTX *ctx = BN_CTX_new();
	BN_CTX_init(ctx);
	
	BN_INIT(rsa->n);
	BN_INIT(rsa->e);
	BN_INIT(rsa->d);
	BN_INIT(rsa->p);
	BN_INIT(rsa->q);
	BN_INIT(rsa->dmp1);
	BN_INIT(rsa->dmq1);
	BN_INIT(rsa->iqmp);
	
	
	BN_set_word( rsa->e, 3ul );
	
	if (P == NULL) {
		
		BN_generate_prime_ex( rsa->p, keybits/2, 1, NULL, NULL, NULL );
	} else {
		BN_hex2bn( &(rsa->p), P );
	}
	
	if (Q == NULL) {
		BN_generate_prime_ex( rsa->q, keybits/2, 1, NULL, NULL, NULL );
	} else {
		BN_hex2bn( &(rsa->q), Q );
	}
	
	BN_mul( rsa->n, rsa->p, rsa->q, ctx );
	
	BIGNUM *pm1 = BN_dup(rsa->p);
	BN_sub(pm1, pm1, BN_value_one());
	
	BIGNUM *qm1 = BN_dup(rsa->q);
	BN_sub(qm1, qm1, BN_value_one());
 	
	BIGNUM * totient = BN_dup(BN_value_one());
	BN_mul(totient, qm1, pm1, ctx);
	
	BN_mod_inverse( rsa->d, rsa->e, totient, ctx );
	
	BN_mod( rsa->dmp1, rsa->d, pm1, ctx );
	BN_mod( rsa->dmq1, rsa->d, qm1, ctx );
	
	BN_mod_inverse( rsa->iqmp, rsa->q, rsa->p, ctx);
		
	BN_clear_free(pm1);
	BN_clear_free(qm1);
	BN_clear_free(totient);
	BN_CTX_free(ctx);
	
	if (RSA_check_key(rsa) != 1 ) {
		printf("ERROR!\n");	
		RSA_free(rsa);
		rsa = NULL;
	}

	
	return rsa;
}

RSA* process_args( int argc, char ** argv ) {
	int retval = 0;
	int opt = -1;
	const char *P = NULL, *Q = NULL;
	int keybits = -1;
	
	while ( (opt = getopt( argc, argv, OPTIONS)) != -1 ) {
		switch(opt) {
			case PRIME_P:
				P = optarg;
				break;
			case PRIME_Q:
				Q = optarg;
				break;
			case KEYBITS:
				keybits = atoi(optarg);
				break;
			case TRAIN_0:
				return training(0);
			case TRAIN_1:
				return training(1);
			case TRAIN_NONE:
				return train_none();
		}
	}
	
	if (keybits == -1 && (P == NULL || Q == NULL)) {
		keybits = DEFAULT_KEY_BITS;
	}
		
	return loadRSA(P, Q, keybits);
}

static RSA * train_none() {
	RSA * rsa = RSA_new();
	BN_CTX *ctx = BN_CTX_new();
	BN_CTX_init(ctx);

	// In this case we only initialize n, e and d	
	BN_INIT(rsa->n);
	BN_INIT(rsa->e);
	BN_INIT(rsa->d);
	BN_INIT(rsa->p);
	BN_INIT(rsa->q);
	
	BN_generate_prime_ex( rsa->p, DEFAULT_KEY_BITS/2, 1, NULL, NULL, NULL );
	BN_generate_prime_ex( rsa->q, DEFAULT_KEY_BITS/2, 1, NULL, NULL, NULL );
	
	BN_mul( rsa->n, rsa->p, rsa->q, ctx );
	
	BN_zero(rsa->d);
	BN_zero(rsa->e);

	BN_CTX_free(ctx);
	
	return rsa;
}

static RSA * training(int zo) {

	RSA * rsa = RSA_new();
	BN_CTX *ctx = BN_CTX_new();
	BN_CTX_init(ctx);

	// In this case we only initialize n, e and d	
	BN_INIT(rsa->n);
	BN_INIT(rsa->e);
	BN_INIT(rsa->d);
	BN_INIT(rsa->p);
	BN_INIT(rsa->q);
	
	BN_generate_prime_ex( rsa->p, DEFAULT_KEY_BITS/2, 1, NULL, NULL, NULL );
	BN_generate_prime_ex( rsa->q, DEFAULT_KEY_BITS/2, 1, NULL, NULL, NULL );
	
	BN_mul( rsa->n, rsa->p, rsa->q, ctx );
	
	BN_one(rsa->d);
	BN_lshift(rsa->d, rsa->d, DEFAULT_KEY_BITS - (zo ^ 0x1));
	if (zo == 1) { // If we are training Zero, set DEFAULT_KEY_BITS bits to 0 and two to 1.
		BN_sub(rsa->d, rsa->d, BN_value_one());
	} else {
		BN_set_bit(rsa->d, 0);
	}
	
	BIGNUM *pm1 = BN_dup(rsa->p);
	BN_sub(pm1, pm1, BN_value_one());
	
	BIGNUM *qm1 = BN_dup(rsa->q);
	BN_sub(qm1, qm1, BN_value_one());
 	
	BIGNUM * totient = BN_dup(BN_value_one());
	BN_mul(totient, qm1, pm1, ctx);
	
	BN_mod_inverse( rsa->e, rsa->d, totient, ctx );
	
	BN_mod( rsa->dmp1, rsa->d, pm1, ctx );
	BN_mod( rsa->dmq1, rsa->d, qm1, ctx );
	
	BN_mod_inverse( rsa->iqmp, rsa->q, rsa->p, ctx);
		
	BN_clear_free(pm1);
	BN_clear_free(qm1);
	BN_clear_free(totient);
	BN_CTX_free(ctx);
	
	if (RSA_check_key(rsa) != 1 ) {
		printf("ERROR!\n");	
		RSA_free(rsa);
		rsa = NULL;
	}
	
	return rsa;
}







