#include <iostream>
#include <string>
#include <cassert>
#include <fstream>
#include <getopt.h>
#include <limits>

#include "codebook.h"
#include "vq.h"
#include "raw2diff.h"
#include "types.h"
#include "raw.h"
#include "pgm.h"
#include "codebook_reduce.h"
#include "comp_fmt.h"


using namespace std;


enum OPT {
	REDUCE = 'r',
	OUTPUT = 'o'
};

const char * USAGE = ""
"classify [-r components.comp] [-o out] vq.cb file\n"
"	Converts raw, pgm or codebook files into listings of\n"
"       labels.\n"
" OUTPUT: \n"
"       file_0.txt [file_n.txt]...\n";

enum { 
	UNKNOWN,
	CB,
	PGM,
	RAW,
};

// Globals:
bool gReduce = false;
std::vector<unsigned int> gComponents;

bool gFill   = false;
raw_t gFillSample;
const int  gFillMax = 10;

std::string gOutfilename;

int determine_type( const std::string & fname ) {
	if ( fname.size() >= 3 ) {
		if (fname.substr(fname.size()-3) == ".cb")
			return CB;
		else if (fname.size() >= 4) {
			if (fname.substr(fname.size()-4) == ".raw")
				return RAW;
			if (fname.substr(fname.size()-4) == ".pgm")
				return PGM;
		}
	}
	return UNKNOWN;
}



void classify_cb( const string & fn_in , const VQ & vq) {
	Codebook cb_in = loadCodebook(fn_in);

	// Change file ending to .txt
	if (gOutfilename == "") {
		gOutfilename = fn_in.substr(0, fn_in.size() -3);
		gOutfilename += ".txt";
	}

	ofstream out(gOutfilename);
	if (!out) {
		cout << "[ERROR] While processing "<< fn_in << ": Could not open file " << gOutfilename <<" for writing." << endl;
		return;
	}

	if (gReduce) {
		cout << "[Warning] Ignoring reduction flag for codebooks. To reduce codebook use tag -r." << endl;
	}
	if (gFill) {
		cout << "[Warning] Will only fill in missing observations for .raw input." << endl;
	}

	for (auto & cw : cb_in ) {
		out << vq.str_encode(cw) << '\n';
	}

	out.close();
	cout << "Classified " << fn_in << " into " << gOutfilename << "." <<  endl;
}

Vector diff2vec( diff_t * data, unsigned int components ) {
	Vector ret(components);
	for (unsigned int i = 0; i < components; i++) {
		ret[i] = data[i];
	}
	return ret;
}

void classify_raw(const string & fn_in , const VQ & vq) {

	// Change file ending to .txt
	if (gOutfilename == "") {
		gOutfilename = fn_in.substr(0, fn_in.size() -4);
		gOutfilename += ".txt";
	}

	ofstream out(gOutfilename);
	if (!out) {
		cout << "[ERROR] While processing "<< fn_in << ": Could not open file " << gOutfilename <<" for writing." << endl;
		return;
	}

	raw_t * rawdata = nullptr;
	diff_t *difdata = nullptr;

	unsigned int b, w, h;

	try {
		raw_load(fn_in.c_str(), nullptr, b, w, h);

		rawdata = new raw_t[w*h];

		raw_load(fn_in.c_str(), rawdata, b, w, h);

		difdata = new diff_t[w*h];

		Vector vec;
		/*raw_t lastTick;
		if (h > 0)
			lastTick = rawdata[0];
		*/
		
		raw2diff(difdata, rawdata, h,  w);
	
		for ( unsigned int i = 0; i < h; i++) {
			/*if (gFill){
				raw_t tick = rawdata[i*w+1];
				const raw_t d_tick = (tick < lastTick ? std::numeric_limits<raw_t>::max() - lastTick + tick : tick - lastTick);
				
				if (d_tick/gFillSample < gFillMax){
					for ( int j = 0; j < d_tick/gFillSample; j++)
						out << '?';
				}
				else {
					out << '\n';
				}

				lastTick = tick;
			}*/

			vec = diff2vec(difdata + w*i, w);
			if (gReduce) {
				vec = reduce_dimensions(vec, gComponents);
			}
 
			const string & temp = vq.str_encode(vec);
			
		    out << temp << '\n';
		}

		out.close();

		cout << "Classified " << fn_in << " into " << gOutfilename << "." <<  endl;
	} catch(...) {
		delete [] rawdata;
		delete [] difdata;
		throw;
	}
}

void classify_pgm(const string & fn_in , const VQ & vq) {

	// Change file ending to .txt
	if (gOutfilename == "") {
		gOutfilename = fn_in.substr(0, fn_in.size() -4);
		gOutfilename += ".txt";
	}

	ofstream out(gOutfilename);
	if (!out) {
		cout << "[ERROR] While processing "<< fn_in << ": Could not open file " << gOutfilename <<" for writing." << endl;
		return;
	}

	if (gFill) {
		cout << "[Warning] Will only fill in missing observations for .raw input." << endl;
	}

	diff_t *difdata = nullptr;

	unsigned int b, w, h;

	try {
		pgm_load(fn_in.c_str(), nullptr, b, w, h);

		difdata = new diff_t[w*h];

		pgm_load(fn_in.c_str(), difdata, b, w, h);


		Vector vec;
		for ( unsigned int i = 0; i < h; i++) {
			vec = diff2vec(difdata + i*w, w);
			if (gReduce) {
				vec = reduce_dimensions(vec, gComponents);
			}
			out << vq.str_encode(vec) << '\n';
		}

		out.close();
		cout << "Classified " << fn_in << " into " << gOutfilename << "." <<  endl;

	} catch(...) {
		delete [] difdata;
		throw;
	}
}

const char * ARGUMENTS = "r:f:";

int main(int argc, char ** argv ) {

	int opt;
	while ((opt = getopt(argc, argv, ARGUMENTS)) != -1) {
		switch(opt) {
			/*case OPT::FILL:
				gFill = true;
				{
					ifstream fin(optarg);
					if (!fin) {
						cout << "[Error] Could not open parameters file \"" << optarg << "\"." << endl;
						return -1;
					}

					//TODO: NB! this has become dependent on experiment 2. 
					string label;
					raw_t value;
					fin >> label >> value; // Ignore first line (OS-TICKS)
					fin >> label >> value; // mod_mul
					gFillSample = value;
					fin >> label >> value; // mod_sqr

					gFillSample += value;
					gFillSample /= 2;

				}
				break;*/
			case OPT::REDUCE:
				gReduce = true;
				gComponents = load_comp(optarg);
				break;
		}
	}



	if (argc-optind < 2 ) {
		cout << "[Error] Missing input files." << endl;
		cout << USAGE << endl;
		return -1;
	} 

	int i = optind;
	if (determine_type(argv[i]) != CB){
		cout << "[Error] First argument not codebook.\n";
		cout << USAGE << endl;
		return -1;
	}

	VQ vq = loadCodebook(argv[i]);
	i++;

	while (i < argc) {
		switch( determine_type(argv[i])) {
			case CB:
				classify_cb(argv[i], vq);
				break;
			case PGM:
				classify_pgm(argv[i], vq);
				break;
			case RAW:
				classify_raw(argv[i], vq);
				break;
			default:
				cout << "[ERROR] Could not process " << argv[i] << ": Fileformat unsupported." << endl;
				break;
		}
		i++;
	}
}
