#include "options.h"
#include "log.h"

#include <iostream>
#include <cstring>
#include <getopt.h>


using namespace std;

const char * USAGE_STR = ""
"USAGE: \n"
"    tag [-aqib] [-v] [-o outfile]  [-r [components-file]] [-T [vq-name]]\n"
"        [-c codebook]... [-L label]... [filename[:label]]... \n"
"\n"
" filename[:label]: Adds all vectors in the given file with the label supplied.\n"
"                   If no label is supplied, the file will be added with the most recent\n"
"                   global label, previously set with -L.\n"
"\n"
"Options:\n"
" -a --append     : Append output to outfile.\n"
" -r --reduce     : Based on the tags, reduce the dimensionality of the input\n"
"                   to only include *significant* components. i.e. components\n"
"                   whose mean differ across labels. If an argument is given\n"
"                   tag will load the component file supplied and uncritically\n"
"                   reduce the input vectors.\n"
" -L --label      : Set global label. Subsequent -f options without a label will use this\n"
"                   global value instead.\n"
" -o --output     : Name of output file, the default is \"default.cb\".\n"
" -v              : Verbose, may also be -vv (very verbose) or -vvv (very very verbose).\n"
" -q --quiet      : Quiet mode, will silently abort on errors.\n"
" -b --batch      : Batch mode, will attempt to continue on errors without prompting.\n"
"                   the user. Implies -q\n"
" -c --combine    : Combine an existing codebook into outfile. \n"
" -T --train-vq   : Creates a VQ using SOM + LVQ. If an argument is supplied, this will\n"
"                   be used as the output name. Otherwise \"default-vq.cb\" is used.\n"
"    --help       : Displays this message.\n"
" -i --info       : Prints statistics about the loaded codebooks.\n"
"\n"
"  VQ Training options \n"
"  --som-nodes=N       : Hints to SOM how many internal nodes to use. The actual nodes used\n"
"                        may be no less than N. (Default 25)\n"
"  --training-rounds=M : As SOM and LVQ are randomized algorithms, subsequent runs may yield\n"
"                        better results. Increase M to try more iterations of the algorithm,\n"
"                        choosing the best result. (Default 1)\n";



enum struct OPT: int  {
	APPEND = 'a',
	COMBINE = 'c',
	HELP = 'h',  // Not available in shortform
	QUIET = 'q',
	REDUCE = 'r',
	TRAIN_VQ = 'T',
	INFO   = 'i',
	CONTINUE_ON_ERROR = 'b',
	VERBOSE= 'v', // Not available in longform
	OUTPUT = 'o',
//	INPUT = 'f',
	LABEL = 'L',
	NO_OUTPUT = 'N',
	// SOM/VQ Options
	SOM_NODES = 0x100,
	TRAINING_ITERATIONS = 0x200,
	
	// END
	UNKNOWN = '?', // For strange enum behaviour
};

const char * shortopts = "aqibvo:f:L:c:r::T::";

struct option longopts[] = {
	{ "append",		no_argument, 		nullptr, (int)OPT::APPEND },
	{ "help", 		no_argument, 		nullptr, (int)OPT::HELP },
	{ "quiet", 		no_argument,	 	nullptr, (int)OPT::QUIET},
	{ "info", 		no_argument,		nullptr, (int)OPT::INFO},
	{ "batch", 		no_argument,		nullptr, (int)OPT::CONTINUE_ON_ERROR},
	{ "no-output",	no_argument, 		nullptr, (int)OPT::NO_OUTPUT},
	{ "output", 	required_argument,	nullptr, (int)OPT::OUTPUT},
//	{ "input-file", required_argument,  nullptr, (int)OPT::INPUT},
	{ "label", 		required_argument,	nullptr, (int)OPT::LABEL},
	{ "combine", 	required_argument, 	nullptr, (int)OPT::COMBINE },
	{ "reduce", 	optional_argument,	nullptr, (int)OPT::REDUCE},
	{ "train-vq", 	optional_argument,	nullptr, (int)OPT::TRAIN_VQ},
	{ "som-nodes",  required_argument,  nullptr, (int)OPT::SOM_NODES},
	{ "training-rounds",  required_argument,  nullptr, (int)OPT::TRAINING_ITERATIONS},

	{ 0, 0, 0, 0 }
};

int process_arguments(int argc, char ** argv ) {
	int longopts_idx = 0;

	int opt = -1;
	const char * active_label = "";
	int retval = 0;
	
	if (argc == 1) {
		cout << USAGE_STR << endl;
		return 1;
	}
	
    while ( (opt = getopt_long(argc, argv, shortopts, longopts, &longopts_idx)) != -1 ) {
        switch((OPT)opt) {
            case OPT::APPEND:
                options.append = true;
                break;
            case OPT::HELP:
                cout << USAGE_STR << endl;
                return 1;
            case OPT::QUIET:
                options.quiet = true;
            	break;
            case OPT::INFO:
                options.info = true;
            	break;
            case OPT::CONTINUE_ON_ERROR:
                options.batch = true;
            	break;
            case OPT::OUTPUT:
                options.out = optarg;
            	break;
            /*case OPT::INPUT:
				{
				    // First, try to split argument
				    const char * fn = strtok(optarg, ":,");
				    const char * label = strtok(nullptr, ":,");
				    
				    if (label == nullptr) label = active_label;
				    
				    options.infiles.emplace( fn, label );
				}
            	break;*/
        	case OPT::COMBINE:
                options.codebooks.emplace(optarg);
        		break;
        	case OPT::LABEL:
        		active_label = optarg;
        		break;
        	case OPT::REDUCE:
        		options.reduce = true;
        		if (optarg) {
	                options.reduce_in = optarg;
        		}
        		break;
        	case OPT::TRAIN_VQ:
                options.create_vq = true;
				if (optarg)
	                options.vq_out = optarg;
	        	break;
	        case OPT::VERBOSE:
	        	incLogLevel();
	        	break;
	        case OPT::SOM_NODES:
	        	options.node_hint = atoi(optarg);
	        	break;
	        case OPT::TRAINING_ITERATIONS:
	        	options.iteration_hint = atoi(optarg);
	        	break;
        	case OPT::UNKNOWN:
        		retval = 1;
        		break;     
        };
    }
    
    for ( int i = optind; i < argc; ++i ) {
		    // First, try to split argument
		    const char * fn = strtok(argv[i], ":,");
		    const char * label = strtok(nullptr, ":,");
		    
		    if (label == nullptr) label = active_label;
		    
		    options.infiles.emplace( fn, label );
    }
    
    return retval;
}

