#include "lvq_helper.h"
#include "lvq.h"
#include "SOM.h"
#include "progress_callback.h"

#include <atomic>
#include <thread>

using namespace std;


static void test_1( const VQ & encoder, const Codebook & training_data, const std::vector<unsigned int> &label_map, const int begin, const int end, atomic<unsigned int> & hits ) {
	for (int i = begin; i < end; i++ ) { 
		auto & elem = training_data[i];
		if (label_map[elem.label()] == encoder.encode(elem) )
			hits++;
	}
}

static unsigned int test(const VQ & encoder, const Codebook & training_data) {
	std::vector<thread> threads;
	atomic<unsigned int> hits(0);

	int nthreads = std::thread::hardware_concurrency();
	int part 	 = training_data.size()/nthreads;

	std::vector<unsigned int> label_map(training_data.getLabelMap().size());
	for ( unsigned int i = 0; i < label_map.size(); i++) {
		label_map[i] = encoder.getLabelId( training_data.getLabelString(i) );
	}

	for ( int i = 0; i < nthreads-1; i++) {
		threads.emplace_back( thread(test_1, std::ref(encoder), std::ref(training_data), std::ref(label_map), i*part, (i+1)*part, std::ref(hits)));
	}
	threads.emplace_back( thread(test_1, std::ref(encoder), std::ref(training_data), std::ref(label_map), (nthreads-1)*part, training_data.size(), std::ref(hits)));

	for ( auto & th : threads ) {
		th.join();
	}
	return hits;
}

void train_som(VQ *candidate, unsigned int *hitrate, atomic<int> & finished, const unsigned int dims, const Codebook & trainingData) {
    SOM som(dims, dims);
	
	som.train(trainingData, 100*trainingData.size());

    *candidate = som.tag_and_purge(trainingData); 
    *hitrate = test(*candidate, trainingData);

	finished = 1;
}

VQ LVQHelper::CreateVQ(const Codebook & trainingData, unsigned int nNodesHint, unsigned int iterationsHint) {

	assert(trainingData.size() < RAND_MAX );

    unsigned int dims = static_cast<unsigned int>(ceil(sqrt(nNodesHint)));
	unsigned int nthreads = min(std::thread::hardware_concurrency(), iterationsHint); 

	VQ candidateVQ[nthreads];
	unsigned int hitrates[nthreads];
	std::thread threads[nthreads];
	atomic<int> flags[nthreads];

    unsigned int bestHitrate = 0, bestlvqHitrate = 0;
    VQ bestVQ;

	int jobs_started = 0;
	int jobs_finished = 0;

	// Start first batch;
	while (jobs_started < nthreads){
		threads[jobs_started] = thread( train_som, candidateVQ + jobs_started, hitrates + jobs_started, ref(flags[jobs_started]), dims, cref(trainingData));
		jobs_started++;
		cout << "Started SOM job " << jobs_started << " of " << iterationsHint << endl;
    }

	while (jobs_finished < iterationsHint) {
		int jobid = -1;
		do {
			for (int job = 0; job < nthreads; job++) {
			   if (flags[job]) {
					jobid = job;
					goto AFTER_LOOP;
			   }
			}
			this_thread::yield();
		} while (true);
AFTER_LOOP:

		threads[jobid].join();
		
		// reset flag:
		flags[jobid] = 0;

		// Check return value.
		if (hitrates[jobid] >= bestHitrate) {
			bestHitrate = hitrates[jobid];
			bestVQ		= std::move(candidateVQ[jobid]);
		}

		jobs_finished++;
		cout << "Finished SOM job: " << jobs_finished << endl;

		if (jobs_started < iterationsHint) {
			threads[jobid] = thread( train_som, candidateVQ + jobid, hitrates + jobid, ref(flags[jobid]), dims, cref(trainingData));
			jobs_started++;
			cout << "Started SOM job " << jobs_started << " of " << iterationsHint << endl;
		}
	}

    
    LVQ bestLVQ(bestVQ);
    
    for (int i = 0; i < iterationsHint; i++ ){
        LVQ temp_lvq(bestVQ);
        
        temp_lvq.train_vq(trainingData, 10000);
        
        unsigned int hitrate = test(temp_lvq, trainingData);
        if ( hitrate >= bestlvqHitrate ) {
        	bestlvqHitrate = hitrate;
        	bestLVQ = std::move(temp_lvq);
        }
    }
    
    cout << "Best hitrates: " << endl;
    cout << "SOM: "<< bestHitrate << "\t" << static_cast<double>(bestHitrate)/trainingData.size() << endl;
    cout << "LVQ: "<< bestlvqHitrate << "\t" << static_cast<double>(bestlvqHitrate)/trainingData.size() << endl;

	return bestLVQ;
}
