#include <iostream>
#include <stdexcept>
#include <string>
#include <cstring>
#include <ctime>
#include <cstdlib>

#include "options.h"
#include "log.h"
#include "loader.h"
#include "lvq_helper.h"
#include "comp_fmt.h"

using namespace std;

int process_arguments(int argc, char ** argv);
int execute();
int print_statistics(const Codebook & );

#define _stringify(x) # x
#define stringify(x) _stringify(x)
#define my_assert(x, msg) if (!(x)) { log( DEBUG, "Assertion failed in " __FILE__ ":" stringify(__LINE__) ": " stringify(x) " " msg ); throw runtime_error(msg); }


int main(int argc, char ** argv) {
	srand(time(0));
	
	int errcode = process_arguments(argc, argv);
	
	if (errcode != 0) 
		return errcode;
			
	/* Check constraints */
	if (options.infiles.size() + options.codebooks.size() == 0 ) {
		log(ERROR, "Error: Missing input files. You must supply at least one input-file or codebook.\n");
		return 1;
	}
	
	
	/*
	
	for (auto elem : options.infiles)
		cout << "Infile: " << elem.first << " tag: " << elem.second << endl;
	for ( auto elem : options.codebooks )
		cout << "Codebook: " << elem << endl;
	cout << "Outfile: " << options.out << endl;
	cout << "Append: " << options.append << endl;
*/	
	
	errcode = execute();
	return errcode;
}


bool prompt() {
	char in = ' ';

	while (in != 'y' && in != 'n') {
		cout << "Y/N: ";
		cin >> in;
		in = tolower(in);
	}

	return in == 'y';
}

bool hasSuffix( const std::string & fn, const char * suffix ) {
	int len_suffix = strlen(suffix);
	int i = fn.size() - len_suffix;
	if ( i >= 0 ) {
		for (int j = 0; j < len_suffix;++j, ++i) {
			if (tolower(fn[i]) != suffix[j] )
				return false;

		}
		return true;
	}
	return false;
}

#if 1

int execute() {
	Loader loader;
	
	if (options.append) {
		try {
			log(VERBOSE, "Attempting to read existing codebook " + options.out + " to append.");
			int n = loader.loadCodebook(options.out);			
			log(VERY_VERBOSE, "Read " + to_string(n) + " codewords.");
		}
		catch (runtime_error & e ) {
			log(VERBOSE, string(e.what()) + "for reading, ignoring.");
		}
		catch (logic_error & e) {
			if (!options.quiet) {
				log(ERROR, string(e.what()) + "\n Continuing may overwrite the file " + options.out + " making it lose its current data. Do you wish to continue? ");
				if (!prompt()) {
					return 1;
				}
			}
		}
	}

	// Loading infiles
	for ( auto elem : options.infiles ) {
		try {
			int vectors = 0;
			log(VERY_VERBOSE, "Processing: " + elem.first + " - " + elem.second + "\n");
			if (hasSuffix(elem.first, ".raw"))
				vectors = loader.loadRaw( elem.first, elem.second );
			else if (hasSuffix(elem.first, ".pgm"))
				vectors = loader.loadPgm( elem.first, elem.second );
			else
				throw std::runtime_error("Unrecognized filetype");

			log(VERBOSE, "Loaded : " + to_string(vectors) + " codewords from " + elem.first + "\n");
		} catch (exception & e) {
			log(ERROR, "Error occured while loading file " + elem.first+ ": " + e.what() + ".");
			if (options.batch) {
				log(ERROR, "(Ignored)\n");
			} else {
				log(ERROR, "\n");
				if (!options.quiet) {
					log(ERROR, "Do you wish to continue? ");
					if (!prompt()) {
						return 1;
					}
				} else {
					return 1;
				}
			}
		}
	}

	// Loading existing codebooks
	for ( auto elem : options.codebooks) {
		try {
			log(VERY_VERBOSE, "Processing: " + elem + "\n");
			int vectors = loader.loadCodebook( elem );
			log(VERBOSE, "Loaded : " + to_string(vectors) + " codewords from " + elem + "\n");
		} catch (exception & e) {
			log(ERROR, "Error occured while loading file " + elem + ": " + e.what() + ".");
			if (options.batch) {
				log(ERROR, "(Ignored)\n");
			} else {
				log(ERROR, "\n");
				if (!options.quiet) {
					log(ERROR, "Do you wish to continue? ");
					if (!prompt()) {
						return 1;
					}
				} else {
					return 1;
				}
			}
		}
	}
	
    // Perform transformations
    if (options.reduce) {
  		log(VERBOSE, "Reducing codebook\n");       
       if ( options.reduce_in != "") {
          	//TODO: Do some errorchecking here.
   			log(VERY_VERBOSE, "Using existing components file: " + options.reduce_in + "\n");       
        	loader.setReducedComponents(load_comp(options.reduce_in));
       }

		
       	loader.reduce();
		log(VERBOSE, "Reduced to %d components.\n", loader.getReducedComponents().size() );
       
       if ( options.reduce_in == "" ) {
       		log(VERY_VERBOSE, "Attempting to save " + options.out + ".comp" + "...");
           	save_comp(options.out + ".comp", loader.getReducedComponents());
           	log(VERY_VERBOSE, " Done\n");
       }
    }

	// Fetch finished codebook.
	Codebook cb_out = loader.combineAll();
	
	// Save reduced codebook.
	while (true) {	
		try {
			log(VERY_VERBOSE, "Attempting to save " + options.out + "...");
			saveCodebook( options.out, cb_out );
			log(VERY_VERBOSE, " Done\n");
			break;
		} catch (exception & e ) {
			log(ERROR, "Error occured while saving file " + options.out + ": " + e.what() + ".\n");
			if (!options.quiet) {
				log(ERROR, "Do you wish to try again? ");
				if (!prompt()) {
					return 1;
				} else {
					continue;
				}
			}
		}
	}
	
	// Create VQ
	if (options.create_vq) {
		VQ res = LVQHelper::CreateVQ(cb_out, options.node_hint, options.iteration_hint);
		
		if (options.info)
			log(INFO, "VQ:");
			print_statistics(res);
		
		while (true) {	
			try {
				log(VERY_VERBOSE, "Attempting to save " + options.vq_out + "...");
				saveCodebook( options.vq_out, res );
				log(VERY_VERBOSE, " Done\n");
				break;
			} catch (exception & e ) {
				log(ERROR, "Error occured while saving file " + options.vq_out + ": " + e.what() + ".\n");
				if (!options.quiet) {
					log(ERROR, "Do you wish to try again? ");
					if (!prompt()) {
						return 1;
					} else {
						continue;
					}
				}
			}
		}
	}
	
	// Output info
	if (options.info)
		log(INFO, "Codebook");
		print_statistics(cb_out);
		

	return 0;
}

#include <iomanip>

int print_statistics(const Codebook & cb_in) {
		cout << "Labels: \n";

		vector<pair<string, int>> labels(cb_in.getLabelMap().size());

		int maxlen = 0;
		for (auto label : cb_in.getLabelMap()) {
			labels[label.second] = make_pair( label.first, 0);
			if (label.first.size() > maxlen) {
				maxlen = label.first.size();
			}
		}

		bool same_n_components = true;
		if ( cb_in.size() != 0) {
			unsigned int components = cb_in[0].size();
			
			for ( auto cw : cb_in ) {
				labels[cw.label()].second++;
				if (cw.size() != components) {
					same_n_components = false;	
				}
			}
		}
		
			
		for ( auto entry : labels ) {
			cout << setw(maxlen + 4) << entry.first << " :  " << entry.second << '\n';
		}
		cout << flush;
		
		log(INFO, "Sanity Check: ");
		if (same_n_components) {
			log(INFO, "Components match.\n");
		} else {
			log(INFO, "Components mismatch.\n");
		}
	return 0;
}
	















#endif
